/*
  Adobe Systems Incorporated(r) Source Code License Agreement
  Copyright(c) 2006 Adobe Systems Incorporated. All rights reserved.

  Please read this Source Code License Agreement carefully before using
  the source code.

  Adobe Systems Incorporated grants to you a perpetual, worldwide,
  non-exclusive, no-charge, royalty-free, irrevocable copyright license,
  to reproduce, prepare derivative works of, publicly display, publicly
  perform, and distribute this source code and such derivative works in
  source or object code form without any attribution requirements.

  The name "Adobe Systems Incorporated" must not be used to endorse or
  promote products derived from the source code without prior written
  permission.

  You agree to indemnify, hold harmless and defend Adobe Systems
  Incorporated from and against any loss, damage, claims or lawsuits,
  including attorney's fees that arise or result from your use or
  distribution of the source code.

  THIS SOURCE CODE IS PROVIDED "AS IS" AND "WITH ALL FAULTS", WITHOUT
  ANY TECHNICAL SUPPORT OR ANY EXPRESSED OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  ALSO, THERE IS NO WARRANTY OF NON-INFRINGEMENT, TITLE OR QUIET
  ENJOYMENT. IN NO EVENT SHALL MACROMEDIA OR ITS SUPPLIERS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE, EVEN
  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// Targets: Adobe Flash Player 9.1 beta 2 for Linux (9.0.21.78)

#include <pulse/pulseaudio.h>
#include <memory.h>
#include <pthread.h>

/* Imported functions */

typedef void* (*T_FPI_Mem_Alloc)(int size);
typedef void (*T_FPI_Mem_Free)(void*);
typedef void (*T_FPI_SoundOutput_FillBuffer)(void* ptr,
                                             char* buffer,
                                             int n_bytes);

struct FPI_Functions
{
    int fpi_count;
    void* fpi_mem_alloc;
    void* fpi_mem_free;
    void* fpi_soundoutput_fillbuffer;
};

/* Exported functions */

void* FPX_Init(void* ptr);

static void FPX_Shutdown(void);

static void* FPX_SoundOutput_Open(void);
static int FPX_SoundOutput_Close(void*);
static int FPX_SoundOutput_Latency(void*);

struct FPX_Functions
{
    int fpx_count;
    void* fpx_shutdown;
    void* fpx_sslsocket_create;
    void* fpx_sslsocket_destroy;
    void* fpx_sslsocket_connect;
    void* fpx_sslsocket_receive;
    void* fpx_sslsocket_send;
    void* fpx_soundoutput_open;
    void* fpx_soundoutput_close;
    void* fpx_soundoutput_latency;
    void* fpx_videoinput_open;
    void* fpx_videoinput_close;
    void* fpx_videoinput_getframe;
};
static struct FPX_Functions fpx_functions;

static T_FPI_Mem_Alloc FPI_Mem_Alloc = 0;
static T_FPI_Mem_Free FPI_Mem_Free = 0;
static T_FPI_SoundOutput_FillBuffer FPI_SoundOutput_FillBuffer = 0;

void* FPX_Init(void* imports)
{
    if (!imports) return 0;

    /* Setup imported functions */
    struct FPI_Functions *fpi_functions = imports;

    if (fpi_functions->fpi_count >= 1)
        FPI_Mem_Alloc = fpi_functions->fpi_mem_alloc;
    if (fpi_functions->fpi_count >= 2)
        FPI_Mem_Free = fpi_functions->fpi_mem_free;
    if (fpi_functions->fpi_count >= 3)
        FPI_SoundOutput_FillBuffer = fpi_functions->fpi_soundoutput_fillbuffer;

    /* Setup exported functions */
    memset(&fpx_functions, 0, sizeof(fpx_functions));
    fpx_functions.fpx_shutdown = FPX_Shutdown;
    fpx_functions.fpx_soundoutput_open = FPX_SoundOutput_Open;
    fpx_functions.fpx_soundoutput_close = FPX_SoundOutput_Close;
    fpx_functions.fpx_soundoutput_latency = FPX_SoundOutput_Latency;
    fpx_functions.fpx_count = 12;

    return (void *)&fpx_functions;
}

static void FPX_Shutdown(void)
{
}

/* Sound support functions */

#define BUFSIZE (4096)

struct output_data {
    pa_threaded_mainloop *mainloop;
    pa_context *context;
    pa_stream *stream;
    uint8_t buf[BUFSIZE];
    pthread_t thread_id;
    int first;
};

static void context_state_cb(pa_context *c, void *userdata) {
    struct output_data *p = userdata;

    assert(c);
    assert(p);

    p->thread_id = pthread_self();
    p->context = c;

    switch (pa_context_get_state(c)) {
        case PA_CONTEXT_READY:
        case PA_CONTEXT_TERMINATED:
        case PA_CONTEXT_FAILED:
            pa_threaded_mainloop_signal(p->mainloop, 0);
            break;

        case PA_CONTEXT_UNCONNECTED:
        case PA_CONTEXT_CONNECTING:
        case PA_CONTEXT_AUTHORIZING:
        case PA_CONTEXT_SETTING_NAME:
            break;
    }
}

static void stream_state_cb(pa_stream *s, void *userdata) {
    struct output_data *p = userdata;

    assert(s);
    assert(p);

    p->thread_id = pthread_self();
    p->stream = s;

    switch (pa_stream_get_state(s)) {

        case PA_STREAM_READY:
        case PA_STREAM_FAILED:
        case PA_STREAM_TERMINATED:
            pa_threaded_mainloop_signal(p->mainloop, 0);
            break;

        case PA_STREAM_UNCONNECTED:
        case PA_STREAM_CREATING:
            break;
    }
}

static void write_data(struct output_data *p) {
    size_t length;

    assert(p);

    /* Wait until timing info is available before we write the second
     * and all subsequent blocks */
    if (!p->first && !pa_stream_get_timing_info(p->stream))
        return;

    length = pa_stream_writable_size(p->stream);

    while (length > 4) {
        size_t l = length;

        if (l > BUFSIZE)
            l = BUFSIZE;

        l &= ~ ((size_t) 3);

        FPI_SoundOutput_FillBuffer(p, (char*) p->buf, l);

        if (pa_stream_write(p->stream, p->buf, l, NULL, 0, PA_SEEK_RELATIVE) < 0)
            break;

        length -= l;

        if (p->first)
            break;
    }

    /* There's no real handling of errors here. Unfortunately the
     * Flash API doesn't export a sane way to do this. With networked
     * audio streams and hotplug-capable audio devices the audio
     * stream might be killed in the middle of nothing, hence it is
     * very unfortunate that we cannot report errors that happen here
     * back to Flash. */

    p->first = 0; /* So, we write the first block noch, remember that */
}

static void stream_request_cb(pa_stream *s, size_t length, void *userdata) {
    struct output_data *p = userdata;

    assert(s);
    assert(length > 0);
    assert(p);

    p->thread_id = pthread_self();

    /* Write some data */
    write_data(p);
}

static void stream_latency_update_cb(pa_stream *s, void *userdata) {
    struct output_data *p = userdata;

    assert(s);
    assert(p);

    p->thread_id = pthread_self();

    /* Try to write some more, in case we delayed the first write until latency data became available */
    write_data(p);
}

static void *FPX_SoundOutput_Open(void) {

    static const pa_sample_spec ss = {
        .format = PA_SAMPLE_S16LE, /* Hmm, Flash wants LE here, not
                                    * NE. This makes porting Flash to
                                    * Big-Endian machines unnecessary
                                    * difficult. */
        .rate = 44100,
        .channels = 2
    };

    struct output_data *p;

    /* Unfortunately we cannot report any useful error string back to
     * Flash. It would be highly preferable if Flash supported some
     * way how we could tell the user what the reason is why audio is
     * not working for him. */

    if (!(p = FPI_Mem_Alloc(sizeof(struct output_data))))
        goto fail;

    memset(p, 0, sizeof(*p));
    p->first = 1;
    p->thread_id = (pthread_t) 0;

    /* First, let's create the main loop */
    if (!(p->mainloop = pa_threaded_mainloop_new()))
        goto fail;

    /* Second, initialize the connection context */
    if (!(p->context = pa_context_new(pa_threaded_mainloop_get_api(p->mainloop), "Adobe Flash")))
        goto fail;

    pa_context_set_state_callback(p->context, context_state_cb, p);

    if (pa_context_connect(p->context, NULL, 0, NULL) < 0)
        goto fail;

    /* Now, let's start the background thread */
    pa_threaded_mainloop_lock(p->mainloop);

    if (pa_threaded_mainloop_start(p->mainloop) < 0)
        goto fail;

    /* Wait until the context is ready */
    pa_threaded_mainloop_wait(p->mainloop);

    if (pa_context_get_state(p->context) != PA_CONTEXT_READY)
        goto unlock_and_fail;

    /* Now, initialize the stream */
    if (!(p->stream = pa_stream_new(p->context, "Flash Animation", &ss, NULL)))
        goto unlock_and_fail;

    pa_stream_set_state_callback(p->stream, stream_state_cb, p);
    pa_stream_set_write_callback(p->stream, stream_request_cb, p);
    pa_stream_set_latency_update_callback(p->stream, stream_latency_update_cb, p);

    if (pa_stream_connect_playback(p->stream, NULL, NULL, PA_STREAM_INTERPOLATE_TIMING|PA_STREAM_AUTO_TIMING_UPDATE, NULL, NULL) < 0)
        goto unlock_and_fail;

    /* Wait until the stream is ready */
    pa_threaded_mainloop_wait(p->mainloop);

    if (pa_stream_get_state(p->stream) != PA_STREAM_READY)
        goto unlock_and_fail;

    pa_threaded_mainloop_unlock(p->mainloop);

    return p;

unlock_and_fail:

    pa_threaded_mainloop_unlock(p->mainloop);

fail:
    if (p)
        FPX_SoundOutput_Close(p);

    return NULL;
}

static int FPX_SoundOutput_Close(void *ptr) {
    struct output_data *p = ptr;

    assert(p);

    if (p->mainloop)
        pa_threaded_mainloop_stop(p->mainloop);

    if (p->stream) {
        pa_stream_disconnect(p->stream);
        pa_stream_unref(p->stream);
    }

    if (p->context) {
        pa_context_disconnect(p->context);
        pa_context_unref(p->context);
    }

    if (p->mainloop)
        pa_threaded_mainloop_free(p->mainloop);

    if (FPI_Mem_Free)
        FPI_Mem_Free(p);

    return 0;
}

static int FPX_SoundOutput_Latency(void *ptr) {
    struct output_data *p = ptr;
    pa_usec_t t = 0;
    int negative;
    int r;

    assert(p);

    /* We lock here only if we are not called from our event loop thread */
    if (!p->thread_id || !pthread_equal(p->thread_id, pthread_self()))
        pa_threaded_mainloop_lock(p->mainloop);

    if (pa_stream_get_latency(p->stream, &t, &negative) < 0 || negative)
        r = 0;
    else
        r = (int) (pa_usec_to_bytes(t, pa_stream_get_sample_spec(p->stream)) >> 2);

    if (!p->thread_id || !pthread_equal(p->thread_id, pthread_self()))
        pa_threaded_mainloop_unlock(p->mainloop);

    return r;
}

